unit OrdNums;
{$WEAKPACKAGEUNIT ON}
{$J-} //no-writeableconst
{$R-} //no-rangechecking
{$Q-} //no-overflowchecking
{.$D-}//no-debug
{
  Fast & (not too) primitive ordinal number conversion
  (byte, shortint, smallint, word, dword, integer, int64)
  strictly speaking this is a 'not real' unit

  Copyright (c) 2004, aa, Inge DR. & Adrian Hafizh.
  Property of PT SOFTINDO Jakarta.
  All rights reserved.

  (this format should stop spammer-bot, to be stripped are:
   at@, brackets[], comma,, overdots., and dash-
   DO NOT strip underscore_)

  mail,to:@[zero_inge]AT@-y.a,h.o.o.@DOTcom,
  mail,to:@[aa]AT@-s.o.f.t,i.n.d.o.@DOTnet
  http://delphi.softindo.net

  Compiler: D5, maybe works also on D4, but not D3 (because of default args value)

  Version: 1.0.0.5
  LastUpdated: 2005.03.11
    added: ROR & ROL ~ why aren't they built-in in the System unit? :(
    add/changed: bin to bins (as in hexs), now better & works also with buffer
    add: octs (bins and hexs complement)
    add/moved from ACommon unit: Blocks (in pure pascal!)

  +collection from other source
    pseudo random generator (Agner Fog, http://www.agner.org) - fixed, enhanced

}
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// The random functions is originated from Uniform Random Number Generator.
// Copyrighted by Agner Fog, http://www.agner.org/random/, licensed under GNU GPL.
// -
// we did a small fix of floating issue, (ought to be?) better initialization
// (at least it really do -what the original version was trying to- with real
// number), enhance double to extended (from full 64-bits wide, not just a
// plain casting), arrange history to directly accomodate int64 result,
// refine min/max range, etc. Delphi's specific.
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// as Neumann says, -Anyone who consider arithmetic means of producing random
// number is, of course, in a state of sin- use this only if you worry about
// different implementation of random function across delphi version
//

interface
uses ACConsts;
//const YES = TRUE;
type
  TInts = ACConsts.TInts;
  TArInts = ACConsts.TArInts;
  TStrs = ACConsts.TStrs;

  TBits = class
  private
    fSize: integer;
    fBits: Pointer;
    procedure Error;
    procedure setSize(Value: integer);
    procedure setBit(Index: integer; Value: Boolean);
    function getBit(Index: integer): Boolean;
  public
    destructor Destroy; override;
    function OpenBit: integer;
    property Bits[Index: integer]: Boolean read getBit write setBit; default;
    property Size: integer read fSize write setSize;
  end;

function setBit(const BitNo, I: integer): integer; overload;
function ResetBit(const BitNo, I: integer): integer; overload;
function ToggleBit(const BitNo, I: integer): integer; overload;
function isBitSet(const BitNo, I: integer): Boolean; overload;

function setBit(const BitNo: integer; const I: int64): int64; overload;
function ResetBit(const BitNo: integer; const I: int64): int64; overload;
function ToggleBit(const BitNo: integer; const I: int64): int64; overload;
function isBitSet(const BitNo: integer; const I: int64): boolean; overload;

// makes a binary (bits) string from a value
function bins(const I: Int64; const BitsWide: integer = 32): string; overload;
function bins(const Buffer: pointer; const BufferLength: integer): string; overload;

// makes an octal string from a value
function octs(const Buffer: pointer; const BufferLength: integer): string; overload;
function octs(const I: int64): string; overload;
function octs(const I: integer): string; overload;

// pretty formatted hexa number
function hexs(const byte: byte; const uppercase: boolean = YES): string; overload;
function hexs(const word: word; const uppercase: boolean = YES): string; overload;
function hexs(const integer: integer; const uppercase: boolean = YES): string; overload;
function hexs(const I: int64; const uppercase: boolean = YES): string; overload;
function hexs(const Buffer: pointer; const BufferLength: integer;
  const Delimiter: Char = #0; const Uppercase: boolean = YES): string; overload;
function hexs(const Buffer: pointer; const BufferLength: integer;
  const Uppercase: boolean; const Delimiter: Char = #0): string; overload;

function IntoHex(const I: integer; const Digits: byte = Sizeof(integer) * 2; UpperCase: boolean = YES): string; register overload;
function IntoHex(const I: Int64; const Digits: byte = Sizeof(int64) * 2; UpperCase: boolean = YES): string; register overload;
//function IntoHex(const I: Int64; const Digits: integer = 0; UpperCase: boolean = YES): string; register overload;
function IntoHex_old(const I: Int64; const Digits: byte = Sizeof(byte)): string; register;

// function intoStr(const I: cardinal; const Digits: integer = 0): string; overload;
// note that the cardinal type, automatically typecasted by the compiler to int64
// since high(cardinal) > high(integer)
// if you are not specifying Digits, its default to Int64's digit size
function IntoStr(const I: integer; const digits: integer = 0): string; register; overload;

// just a speedy wrapper for Inttostr, IntToHex and StrToInt
// IsHex is just an auto prepend '$' hex-specifier,
// so do not call IsHex = YES, if the S has already had '$'
function IntoStr(const I: int64; const digits: integer = 0): string; register; overload;
function IntOf(const S: string; const DefaultValue: integer = 0): integer; register; overload;
//function IntOf(const S: string; const IsHex: Boolean = FALSE; const DefaultValue: integer = 0): integer; overload;
//function Int64Of(const S: string; const IsHex: Boolean = FALSE; const DefaultValue: Int64 = 0): Int64; overload;
//function IntOf(const S: string; const DefaultValue: integer; const IsHex: Boolean = FALSE): integer; overload;
//function Int64Of(const S: string; const DefaultValue: Int64; const IsHex: Boolean = FALSE): Int64; overload;

// much simpler StrToIntDef clones, do not expect too much
function Str2Int(const S: string; const IsHex: Boolean = FALSE; const DefaultValue: integer = 0): integer; overload;
function Str2Int64(const S: string; const IsHex: Boolean = FALSE; const DefaultValue: Int64 = 0): Int64; overload;
function Str2Int(const S: string; const DefaultValue: integer; const IsHex: Boolean = FALSE): integer; overload;
function Str2Int64(const S: string; const DefaultValue: Int64; const IsHex: Boolean = FALSE): Int64; overload;

function Min(const a, b: integer): integer; register overload
function Max(const a, b: integer): integer; register overload
//function UMin(const a, b: integer): integer; register overload
//function UMax(const a, b: integer): integer; register overload
function Min(const a, b: cardinal): cardinal; register overload //unsigned integer
function Max(const a, b: cardinal): cardinal; register overload //unsigned integer
function uMin(const a, b: Int64): Int64; register; overload //unsigned int64
function uMax(const a, b: Int64): Int64; register; overload //unsigned int64

// ROL & ROR family, which should have been included in the System unit
function rol(const I: integer): integer; register overload;
function ror(const I: integer): integer; register overload;
function rol(const I: int64): int64; register overload;
function ror(const I: int64): int64; register overload;

function rol(const I: integer; const ShiftCount: integer): integer; register overload;
function ror(const I: integer; const ShiftCount: integer): integer; register overload;
function rol(const I: int64; const ShiftCount: integer): int64; register overload;
function ror(const I: int64; const ShiftCount: integer): int64; register overload;

// and why aren't shlr & shld either? :)
procedure shld(var A, B: integer); register overload;
procedure shld(var A, B: integer; const ShiftCount: integer); register overload;
procedure shld(var A, B: int64); register overload;
procedure shld(var A, B: int64; const ShiftCount: integer); register overload;

procedure shrd(var A, B: integer); register overload;
procedure shrd(var A, B: integer; const ShiftCount: integer); register overload;
procedure shrd(var A, B: int64); register overload;
procedure shrd(var A, B: int64; const ShiftCount: integer); register overload;

// note:
//   the optimization by the compiler will lead to integer version of ROL/ROR
//   if the value of I is within the range of integer (you should typecasted it to Int64).
//
//  // priorly ShiftCount is of type byte, now changed to integer,
//  // this note is no longer applicable; this only for remainder,
//  // on the similar circumtance, this comment is still valid though.
//  // obsolete:  BEWARE if you (on slippery) supplied an integer type
//  // obsolete:  (NOT byte) of ShiftCount, EVEN if I is type of int64,
//  // obsolete:  the compiler will take an integer version of ROL/ROR.
//  // obsolete:  (you MUST explicitly typecasted it to byte)
//
//   to avoid confusion, rol for int64 idname might better be changed to (ie.) rol64 instead
//

// block string formatting, distribute string in blocks of BlockLen length,
// customizable block length and delimiter, leftwise or rightwise
//
// e.g. 1234567890 -> 123 456 789 0  (length = 3 (default), delim = A SPACE, leftwise)
//      1234567890 -> 1 234 567 890  (length = 3 (default), delim = A SPACE, righwise)
//
//   Prefix and Suffix length are number of firstly/lastly characters which should be ignored
//     in formatting
//
// to format Buffer as hex use Hexs instead!
function Blocks(const S: string; const delimiter: string = SPACE; const BlockLen: integer = 3;
  const LeftWise: Boolean = FALSE; //const Interval: integer = 1;
  const SkipPrefixLength: integer = 0; const SkipSuffixLength: integer = 0): string; overload;
function Blocks(const I: integer; const delimiter: string = SPACE; const BlockLen: integer = 3;
  const LeftWise: Boolean = FALSE; //const Interval: integer = 1;
  const SkipPrefixLength: integer = 0; const SkipSuffixLength: integer = 0): string; overload;

function RandomInt64: int64; overload
function RandomInt(const Max: cardinal): cardinal; register overload
function RandomInt(const Min, Max: integer): integer; register overload
procedure RandomInit(const I: integer = __AAMAGIC0__); register

const
  RandseedEx: array[0..4] of integer = (__AAMAGIC0__, __AAMAGIC1__, __AAMAGIC2__, __CRC32Poly__, -1);

const
  BitsperByte = 8;
  BitsperWord = SizeOf(word) * BitsPerByte;
  BitsperInt = SizeOf(integer) * BitsPerByte;
  BitsperInt64 = SizeOf(int64) * BitsPerByte;

{$IFDEF TESTA_57}
  // (currently unused, reciprocal int64 / 10)
  MaxInt64 = high(int64); //9223372036854775807 *2

  MaxUInt64 = 18446744073709551615.0 + 1; // round error adjustment bit-1
  Int64e00 = round(MaxUInt64 / 1E00); //18 446 744 073 709 551 615;
  Int64e01 = round(MaxUInt64 / 1E01); // 1 844 674 407 370 955 162;
  Int64e02 = round(MaxUInt64 / 1E02); //   184 467 440 737 095 516;
  Int64e03 = round(MaxUInt64 / 1E03); //    18 446 744 073 709 552;
  Int64e04 = round(MaxUInt64 / 1E04); //     1 844 674 407 370 955;
  Int64e05 = round(MaxUInt64 / 1E05); //       184 467 440 737 096;
  Int64e06 = round(MaxUInt64 / 1E06); //        18 446 744 073 710;
  Int64e07 = round(MaxUInt64 / 1E07); //         1 844 674 407 371;
  Int64e08 = round(MaxUInt64 / 1E08); //           184 467 440 737;
  Int64e09 = round(MaxUInt64 / 1E09); //            18 446 744 074;
  Int64e10 = round(MaxUInt64 / 1E10); //             1 844 674 407;
  Int64e11 = round(MaxUInt64 / 1E11); //               184 467 441;
  Int64e12 = round(MaxUInt64 / 1E12); //                18 446 744;
  Int64e13 = round(MaxUInt64 / 1E13); //                 1 844 674;
  Int64e14 = round(MaxUInt64 / 1E14); //                   184 467;
  Int64e15 = round(MaxUInt64 / 1E15); //                    18 447;
  Int64e16 = round(MaxUInt64 / 1E16); //                     1 845;
  Int64e17 = round(MaxUInt64 / 1E17); //                       184;
  Int64e18 = round(MaxUInt64 / 1E18); //                        18;
  Int64e19 = round(MaxUInt64 / 1E19); //                         2;

  KInt64ext: array[0..19] of int64 = (
    Int64e00, Int64e01, Int64e02, Int64e03, Int64e04,
    Int64e05, Int64e06, Int64e07, Int64e08, Int64e09,
    Int64e10, Int64e11, Int64e12, Int64e13, Int64e14,
    Int64e15, Int64e16, Int64e17, Int64e18, Int64e19
    );

  Int64e00Lo = cardinal(Int64e00);
  Int64e01Lo = cardinal(Int64e01);
  Int64e02Lo = cardinal(Int64e02);
  Int64e03Lo = cardinal(Int64e03);
  Int64e04Lo = cardinal(Int64e04);
  Int64e05Lo = cardinal(Int64e05);
  Int64e06Lo = cardinal(Int64e06);
  Int64e07Lo = cardinal(Int64e07);
  Int64e08Lo = cardinal(Int64e08);
  Int64e09Lo = cardinal(Int64e09);
  Int64e10Lo = cardinal(Int64e10);
  Int64e11Lo = cardinal(Int64e11);
  Int64e12Lo = cardinal(Int64e12);
  Int64e13Lo = cardinal(Int64e13);
  Int64e14Lo = cardinal(Int64e14);
  Int64e15Lo = cardinal(Int64e15);
  Int64e16Lo = cardinal(Int64e16);
  Int64e17Lo = cardinal(Int64e17);
  Int64e18Lo = cardinal(Int64e18);
  Int64e19Lo = cardinal(Int64e19);

  KInt64extLo: array[0..19] of cardinal = (
    Int64e00Lo, Int64e01Lo, Int64e02Lo, Int64e03Lo, Int64e04Lo,
    Int64e05Lo, Int64e06Lo, Int64e07Lo, Int64e08Lo, Int64e09Lo,
    Int64e10Lo, Int64e11Lo, Int64e12Lo, Int64e13Lo, Int64e14Lo,
    Int64e15Lo, Int64e16Lo, Int64e17Lo, Int64e18Lo, Int64e19Lo
    );

  Int64e00Hi = cardinal(Int64e00 shr 32);
  Int64e01Hi = cardinal(Int64e01 shr 32);
  Int64e02Hi = cardinal(Int64e02 shr 32);
  Int64e03Hi = cardinal(Int64e03 shr 32);
  Int64e04Hi = cardinal(Int64e04 shr 32);
  Int64e05Hi = cardinal(Int64e05 shr 32);
  Int64e06Hi = cardinal(Int64e06 shr 32);
  Int64e07Hi = cardinal(Int64e07 shr 32);
  Int64e08Hi = cardinal(Int64e08 shr 32);
  Int64e09Hi = cardinal(Int64e09 shr 32);
  Int64e10Hi = cardinal(Int64e10 shr 32);
  Int64e11Hi = cardinal(Int64e11 shr 32);
  Int64e12Hi = cardinal(Int64e12 shr 32);
  Int64e13Hi = cardinal(Int64e13 shr 32);
  Int64e14Hi = cardinal(Int64e14 shr 32);
  Int64e15Hi = cardinal(Int64e15 shr 32);
  Int64e16Hi = cardinal(Int64e16 shr 32);
  Int64e17Hi = cardinal(Int64e17 shr 32);
  Int64e18Hi = cardinal(Int64e18 shr 32);
  Int64e19Hi = cardinal(Int64e19 shr 32);

  KInt64extHi: array[0..19] of cardinal = (
    Int64e00Hi, Int64e01Hi, Int64e02Hi, Int64e03Hi, Int64e04Hi,
    Int64e05Hi, Int64e06Hi, Int64e07Hi, Int64e08Hi, Int64e09Hi,
    Int64e10Hi, Int64e11Hi, Int64e12Hi, Int64e13Hi, Int64e14Hi,
    Int64e15Hi, Int64e16Hi, Int64e17Hi, Int64e18Hi, Int64e19Hi
    );
{$ENDIF TESTA_57}

// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// private debug/test only, DO NOT use!
// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
type
  TIDStr = class
  private
    fInts: TInts;
    fStrs: TStrs;
    fMaxID: integer;
    fCount: integer;
    function getValue(const ID: integer): string;
    procedure setValue(const ID: integer; const Value: string);
    procedure setCount(const Value: integer);
  public
    destructor Destroy; override;
    constructor Create; overload;
    constructor Create(const Size: integer); overload;
    property Strings[const ID: integer]: string read getValue write setValue; default;
    function IDOf(const Value: string): integer;
    function IDAtPos(const Pos: integer): integer; // ordinal position (0-based)
    function ValueAtPos(const Pos: integer): string; // ordinal position (0-based)
    property Count: integer read fCount write setCount;
    function delete(const ID: integer): boolean;
    procedure join(const IntStr: TIDStr);
    procedure Clear;
  end;

implementation
//uses MinMaxMid

//type
//  TBinaryType (binBoolean, binByte, binWord, binDword, binInt64
type
  r64 = packed record
    Lo, hi: integer;
  end;

// function _fillzero(const S: string; const digits: integer; const positive: boolean): string;
// // dont ever let anything critical called under the try..except/finally construct
// // for that would degrade the performance down, the compiler inserts fs: everywhere
// var
//   n: integer;
// begin
//   n := length(S);
//   if digits <= n then Result := S
//   else begin
//     if positive then
//       Result := StringOfchar(zero, digits - n) + S
//     else
//       Result := dash + StringOfChar(zero, digits - n) + copy(S, 2, n);
//   end;
// end;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Notes:
//   Min-Max functions excerpted from unit MinMaxMid
//   (produced by the same authors)
//   please get the recent version
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

function Min(const a, b: integer): integer; overload asm
   cmp a, b; jle @end
     mov a, b
   @end:
end;

function Max(const a, b: integer): integer; overload asm
   cmp a, b; jge @end
     mov a, b
   @end:
end;

function Min(const a, b: cardinal): cardinal; overload asm
  cmp a, b; jbe @@done
    mov a, b
  @@done:
end;

function Max(const a, b: cardinal): cardinal; overload asm
  cmp a, b; jae @@done
  mov a, b
  @@done:
end;

function uMin(const a, b: int64): int64; overload asm
  cmp edx, b.r64.hi; jbe @@end
  cmp eax, b.r64.Lo; jbe @@end
  mov eax, b.r64.Lo
  mov edx, b.r64.hi
  @@end:
end;

function uMax(const a, b: int64): int64; overload asm
  cmp edx, b.r64.hi; jae @@end
  cmp eax, b.r64.Lo; jae @@end
  mov eax, b.r64.Lo
  mov edx, b.r64.hi
  @@end:
end;

function Min(const a, b: int64): int64; overload asm
  cmp edx, b.r64.hi; jle @@end
  cmp eax, b.r64.Lo; jbe @@end
  mov eax, b.r64.Lo
  mov edx, b.r64.hi
  @@end:
end;

function Max(const a, b: int64): int64; overload asm
  cmp edx, b.r64.hi; jge @@end
  cmp eax, b.r64.Lo; jae @@end
  mov eax, b.r64.Lo
  mov edx, b.r64.hi
  @@end:
end;

//~~~~~~~~~~~~~~~~~~~~~~
// end minmax functions
//~~~~~~~~~~~~~~~~~~~~~~

//~~~~~~~~~~~~~~~~~~~~~~
// ROL/ROR
//~~~~~~~~~~~~~~~~~~~~~~

function rol(const I: integer): integer; overload asm rol I, 1 end;
function rol(const I: int64): int64; overload asm
    mov edx, I.r64.hi  // using register is faster than directly accessing memory
    mov eax, I.r64.lo // in Pentium they could also be run parallelized
    shl eax, 1; rcl edx, 1
    jnc @done; or eax, 1
  @done: //popfd
end;
function ror(const I: integer): integer; overload asm ror I, 1 end;
function ror(const I: int64): int64; overload asm
    mov edx, I.r64.hi  // using register is faster than directly accessing memory
    mov eax, I.r64.lo //  in Pentium they could also be run parallelized
    shr edx, 1; rcr eax, 1
    jnc @done; or edx, 1 shl 31
  @done: //popfd
end;
function rol(const I: int64; const ShiftCount: integer): int64; register overload asm
    mov ecx, ShiftCount // as Intel says, upon shift this value will be taken MODULO 32
    mov edx, I.r64.hi  // using register is faster than directly accessing memory
    mov eax, I.r64.lo // in Pentium they could also be run parallelized
    and ecx, $3f; jz @exit
    cmp cl, 32; jb @begin
    //xchg eax, edx   // avoid LOCK prefixed xchg instruction
    mov eax, edx      // simple move should be faster & pairing enable
    mov edx, I.r64.lo //
    jz @exit
  @begin:
    push ebx; mov ebx, eax
    shld eax, edx, cl
    shld edx, ebx, cl
  @done: pop ebx
  @exit:
end;
function ror(const I: int64; const ShiftCount: integer): int64; register overload asm
    mov ecx, ShiftCount // as Intel says, upon shift this value will be taken MODULO 32
    mov edx, I.r64.hi  // using register is faster than directly accessing memory
    mov eax, I.r64.lo // in Pentium they could also be run parallelized
    and ecx, $3f; jz @exit
    cmp cl, 32; jb @begin
    //xchg eax, edx   // avoid LOCK prefixed xchg instruction
    mov eax, edx      // simple move should be faster & pairing enable
    mov edx, I.r64.lo //
    jz @exit
  @begin:
    push ebx; mov ebx, edx
    shrd edx, eax, cl
    shrd eax, ebx, cl
  @done: pop ebx
  @exit:
end;
function rol(const I: integer; const ShiftCount: integer): integer; register overload asm mov ecx, ShiftCount; rol I, cl end;
function ror(const I: integer; const ShiftCount: integer): integer; register overload asm mov ecx, ShiftCount; ror I, cl end;

procedure shld(var A, B: integer); register overload asm mov edx, B; shld A, edx, 1; mov B, edx end;
procedure shld(var A, B: integer; const ShiftCount: integer); register overload asm mov edx, B; shld A, edx, cl; mov B, edx end;

procedure shld(var A, B: int64); register overload asm
    mov edx, B.r64.hi
    mov eax, B.r64.lo
    shl eax, 1; rcl edx, 1
    jnc @done1; or eax, 1
  @done1: //popfd
    mov B.r64.hi, edx
    mov B.r64.lo, eax
    mov edx, A.r64.hi
    mov eax, A.r64.lo
    shl eax, 1; rcl edx, 1
    mov A.r64.hi, edx
    mov A.r64.lo, eax
  @done: //popfd
end;

procedure shld(var A, B: int64; const ShiftCount: integer); register overload asm
end;

procedure shrd(var A, B: integer); register overload asm mov edx, B; shrd A, edx, 1; mov B, edx end;
procedure shrd(var A, B: integer; const ShiftCount: integer); register overload asm mov edx, B; shrd A, edx, cl; mov B, edx end;
procedure shrd(var A, B: int64); register overload asm end;
procedure shrd(var A, B: int64; const ShiftCount: integer); register overload asm end;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Blocks. Moved in from ACommon unit
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function Blocks(const S: string; const delimiter: string = SPACE; const BlockLen: integer = 3;
  const LeftWise: Boolean = FALSE; //const Interval: integer = 1;
  const SkipPrefixLength: integer = 0; const SkipSuffixLength: integer = 0): string; overload;
const
  MINBLOCKLEN: integer = 1;
var
  i, j, k, l, r: integer;
  prefix, suffix: string;
begin
  l := length(S);
  k := SkipPrefixLength + SkipSuffixLength;
  if (delimiter = '') or (l <= 1) or (SkipPrefixLength < 0) or (SkipSuffixLength < 0) or (l - k <= 1) then
    Result := S
  else begin
    if SkipPrefixLength > 0 then prefix := copy(S, 1, SkipPrefixLength);
    if SkipSuffixLength > 0 then suffix := copy(S, l - k + 1, SkipSuffixLength);
    Result := copy(S, SkipPrefixLength + 1, l - k);
    j := Max(word(BlockLen), MINBLOCKLEN); // max.65535

    l := length(Result);
    i := l div j;
    if (l mod j = 0) then dec(i);

    if LeftWise then
      for r := i downto 1 do
        insert(delimiter, Result, (r * j) + 1)
    else begin
      for r := i downto 1 do
        insert(delimiter, Result, length(Result) - (r * j) + 1);
    end;

    if SkipPrefixLength > 0 then Result := prefix + Result;
    if SkipSuffixLength > 0 then Result := Result + suffix;
  end;
end;

function Blocks(const I: integer; const delimiter: string = SPACE; const BlockLen: integer = 3;
  const LeftWise: Boolean = FALSE; //const Interval: integer = 1;
  const SkipPrefixLength: integer = 0; const SkipSuffixLength: integer = 0): string; overload;
begin
  Result := blocks(IntoStr(I), delimiter, BlockLen, LeftWise, SkipPrefixLength, SkipSuffixLength);
end;

// Simple intoStr
// function intoStr(const I: integer; const digits: integer = 0): string;
// const zero = '0'; dash = '-';
// var
//   n: integer;
// begin
//   if I = 0 then Result := StringOfChar(zero, max(1, digits))
//   else begin
//     Str(I: 0, Result);
//     n := length(Result);
//     if digits > n then begin
//       if i > 0 then
//         Result := StringOfchar(zero, digits - n) + Result
//       else
//         Result := dash + StringOfChar(zero, digits - n) + copy(Result, 2, n);
//     end;
//   end;
// end;

const zero = CHAR_ZERO; //'0'; //
const dash = CHAR_DASH; //'-'; //

function intoStr(const I: integer; const Digits: integer = 0): string; overload;
var
  Value: string[20];
  minus: char;
asm
  test eax, eax; jnz @goon
    mov eax, 1
    call Max
    and eax, $ff
    mov edx, eax; //mov ecx, Result // already done
    mov eax, zero
    call System.@LStrOfchar
  jmp @end
  @goon:
    PUSH Result; PUSH Digits
    mov minus, zero
    or eax, eax; jge @skipmin
    //setl minus
    mov minus, dash;
    @skipmin:
    xor edx, edx; lea ecx, Value
    call System.@StrLong
    pop Digits; //pop Result
    mov al, byte ptr Value
    and eax, $1f
    cmp Digits, eax; jg @moredigits
      pop eax {Result}; lea edx, Value
      call System.@LStrFromString

      jmp @end

    @moredigits:
      //mov edx, eax; //mov ecx, Result // already done
      mov Result, [esp]; mov eax, zero;
      call System.@LStrOfchar
      xor ecx, ecx
      mov cl, byte ptr[Value]
      mov dl, minus
      pop eax {Result}; mov eax, [eax]; mov [eax], dl
      add eax, [eax-4]; dec eax  // end of Result
      cmp dl, dash; lea edx, Value + ecx
      sete ch; sub cl, ch
      @loop:
        mov ch, [edx]; lea edx, edx-1
        mov [eax], ch; lea eax, eax-1
        dec cl
      jne @loop

  jmp @end
  @end:
end;

// Simple intoStr
// function intoStr(const I: integer; const digits: integer = 0): string;
// const zero = '0'; dash = '-';
// var
//   n: integer;
// begin
//   if I = 0 then Result := StringOfChar(zero, max(1, digits))
//   else begin
//     Str(I: 0, Result);
//     n := length(Result);
//     if digits > n then begin
//       if i > 0 then
//         Result := StringOfchar(zero, digits - n) + Result
//       else
//         Result := dash + StringOfChar(zero, digits - n) + copy(Result, 2, n);
//     end;
//   end;
// end;

function intoStr64(const I: int64; const Digits: integer = 0): string; overload;
const zero = CHAR_ZERO; //'0'; //
const dash = CHAR_DASH; //'-'; //
var
  n: integer;
begin
  Str(I: 0, Result);
  n := length(Result);
  if digits > n then begin
    if I > 0 then
      Result := StringOfchar(zero, digits - n) + Result
    else
      Result := dash + StringOfChar(zero, digits - n) + Copy(Result, 2, n);
  end;
end;

function intoStr(const I: int64; const Digits: integer = 0): string; overload asm
// Int64 always passed via stack, here eax = Digits and edx = @Result
// note: cardinal type will be treated as Int64 (only here?)
    mov ecx, r64(I).hi
    test ecx, ecx; jz @testCard0 // MSB = 0, test for cardinality
    cmp ecx, -1; jne @I64 // if MSB = -1, test for sign-extended
  @testSignex:
    mov ecx, r64(I).lo
    test ecx, ecx; js @I32  // if LSB also negative, call faster I32 instead
    jmp @I64
  @testCard0:
    mov ecx, r64(I).lo
    test ecx, ecx; jns @I32 // if LSB positive, call faster I32 instead
    //jmp @I64
  @I64:
    PUSH r64(I).hi; PUSH r64(I).lo
    call intoStr64
    jmp @end
  @I32:
    mov ecx, Result
    mov edx, Digits
    mov eax, r64(I).lo;
    call intoStr
  @end:
end;

function IntOf(const S: string; const DefaultValue: integer = 0): integer; // lightweight version
var e: integer;
begin
  val(S, Result, e);
  if e <> 0 then
    Result := DefaultValue;
end;

function Intf(const S: string; const DefaultValue: integer = 0): integer; // lightweight version
var
  d: double;
  e: integer;
begin
  val(S, d, e);
  if e <> 0 then
    Result := DefaultValue
  else
    Result := trunc(d)
end;

function Intf64(const S: string; const DefaultValue: integer = 0): int64; // lightweight version
var
  d: double;
  e: integer;
begin
  val(S, d, e);
  if e <> 0 then
    Result := DefaultValue
  else
    Result := trunc(d)
end;

function Str2Int64(const S: string; const IsHex: Boolean = FALSE; const DefaultValue: Int64 = 0): Int64; overload;
var
  e: integer;
begin
  if IsHex then begin
    if (S <> '') and (S[1] <> '$') then
      Val('$0' + S, Result, e)
    else
      Val(S, Result, e)
  end
  else
    Val(S, Result, e);
  if e <> 0 then
    Result := DefaultValue;
end;

function Str2Int(const S: string; const IsHex: Boolean = FALSE; const DefaultValue: integer = 0): integer; overload;
var
  e: integer;
begin
  //Result := StringToInt64(S, IsHex, DefaultValue);
  if IsHex then
    Val('$0' + S, Result, e)
  else
    Val(S, Result, e);
  if e <> 0 then
    Result := DefaultValue;
end;

function Str2Int(const S: string; const DefaultValue: integer; const IsHex: Boolean = FALSE): integer; overload;
begin
  Result := Str2Int(S, isHex, DefaultValue);
end;

function Str2Int64(const S: string; const DefaultValue: Int64; const IsHex: Boolean = FALSE): Int64; overload;
begin
  Result := Str2Int64(S, isHex, DefaultValue);
end;

function Digitize(const S: string; const Digits: integer; Negative: boolean = FALSE; UpperCase: boolean = YES): string;
//function Min(a, b: integer): integer; asm
//  cmp a, b; jle @end
//    mov a, b
//  @end:
//  end;
type
  TNegative = boolean;
  TUpperCase = boolean;
const
  ef = 'f';
  zero = CHAR_ZERO;
  space = CHAR_SPACE;
  UPPING = not ord(space);
  fills: array[TUpperCase, TNegative] of char = ((zero, ef), (zero, char(ord(ef) and UPPING)));

  function firstNonZero: integer;
  var
    fill: char;
  begin
    if S = '' then
      Result := 0
    else if Negative then
      Result := 1
    else begin
      fill := fills[UpperCase, Negative];
      for Result := 1 to length(S) do
        if S[Result] <> fill then
          break
    end;
    //debug:
    //if Result = 0 then
    //  Result := 1;
  end;

var
  L: integer;
begin
  if (Digits < 1) then
    Result := S
  else begin
    L := length(S);
    if Digits = L then
      Result := S
    else if Digits < L then
      Result := copy(S, min(L - Digits + 1, firstNonZero), L)
    else
      Result := stringofchar(fills[UpperCase, Negative], Digits - L) + S;
  end;
end;

function IntoHex(const I: integer; const Digits: byte = Sizeof(integer) * 2; UpperCase: boolean = YES): string; register overload;
const
  MaxByte = high(byte);
  MaxWord = high(word);
  //MinShort = low(shortint);  MinSmall = low(SmallInt);
begin
  case i of
    //allowing negative values for byte & word type are rather confusing than useful
    {MinShort}0..MaxByte: Result := hexs(byte(i), UpperCase);
    {MinSmall..MinShort - 1,}Maxbyte + 1..maxword: Result := hexs(word(i), UpperCase);
    //-maxint..(-maxword - 1), maxword + 1..maxint: Result := hexs(i, UpperCase);
  else
    Result := hexs(i, UpperCase);
  end;
  Result := Digitize(Result, Digits, i < 0, UpperCase);
end;

function IntoHex(const I: Int64; const Digits: byte = Sizeof(int64) * 2; UpperCase: boolean = YES): string; register overload;
begin
  //if r64(I).hi<>0 then
  if (I < Low(integer)) or (I > high(integer)) then
    Result := Digitize(hexs(i, UpperCase), Digits, i < 0, UpperCase)
  else
    Result := intoHex(integer(i), Digits, UpperCase)
end;

// function IntoHex_notused(const I: Int64): string; register; overload;
// var
//   S: shortstring;
// asm
// end;
//

function IntoHex_old(const I: Int64; const Digits: byte = Sizeof(byte)): string; register;
// Digits is number of HEXADECIMAL chars representation, should be multiple of 2
// example: IntoHex(1234, 5) = 0004D2 (6 chars anyway)
// This routine actually is obsoleted, used for simple conversion only
const
  DIGITSQUAD = Sizeof(Int64); // 8 byte
var
  S: shortstring;
  widow: boolean;
begin
  asm
    @@Start:
      PUSH esi; PUSH edi
      PUSH Result
      lea esi, I
      and Digits, 11111b
      inc Digits; shr Digits, 1

      mov al, Digits
      and eax, 0FFh
      //mov al, HexDigits
      //and eax, 0ffh
      mov widow, false
      mov edi, DIGITSQUAD
      cmp eax, edi; ja @_checkdone

      @_checkdigit:
      mov ecx, 4; mov edx, [esi+4] //I.hi

      @_Loop1: rol edx, 8
      or dl, dl; jnz @_checkdone
      dec edi; cmp edi, eax; jb @_recall
      dec ecx; jnz @_Loop1

      mov edx, [esi]
      @_Loop2: rol edx, 8
      or dl, dl; jnz @_checkdone
      dec edi; jz @_recall
      cmp edi, eax; jge @_Loop2

      @_recall: inc edi
      @_checkdone:

      mov eax, edi; lea edi, S

      mov ecx, eax; add esi, eax
      dec esi; shl eax, 1
      cld; stosb

    @@Loop:
      std; lodsb
      mov ah, al; shr al, 04h
      add al, 90h; daa
      adc al, 40h; daa
      cld; stosb
      mov al, ah; and al, 0Fh
      add al, 90h; daa
      adc al, 40h; daa
      stosb
      dec ecx; jnz @@Loop

      lea edx, S; pop eax
      call System.@LStrFromString
      pop edi; pop esi
    @@Stop:
  end;
end;

//System
//procedure _StrLong(Val, Width: Longint; var S: ShortString);
//procedure _LStrSetLength{var Str: ANSIString; NewLength: integer};
//procedure _LStrFromString(var Dest: ANSIString; const Source: ShortString);
//procedure _LStrOfCharCc: Char; count: integer): ANSIString;

//procedure  _NewAnsiString{Length: Longint};

//procedure UniqueString(var Str: string);
//procedure _LStrAsg{var Dest: ANSIString; Source: ANSIString};
//procedure _LStrLAsg{var Dest: ANSIString; Source: ANSIString};

function bins_OK(const Buffer: pointer; const BufferLength: integer): string; register overload assembler asm
  @@Start:
    or eax, eax; jz @@Stop     // insanity checks
    or edx, edx; jg @@Begin    // can not exceed 2GB
    xor eax, eax
    jmp @@Stop
  @@Begin:
    push esi; push edi; push ebx
    mov esi, Buffer
    // edx (as well as ecx), usually destroyed by Str operation
    mov ebx, edx               // BufLen
    mov eax, Result            // where the result will be stored
    call System.@LStrClr       // cleared for ease
    //shl edx, 3
    lea edx, [ebx*8]           // BufferLength * 8 bits
    call System.@LStrSetLength // result: new allocation pointer in EAX
    mov edi, [eax]               // put new allocated ptr of @Result to edi

    push edi

    lea ecx, [ebx*2]
    mov eax, '0000'; rep stosd

    mov edi, [esp]
    //mov ecx, ebx               // get bufferlength back to ecx
    mov Ch, '1'                  // buffer fillin value

  @@Loop: lodsb; mov cl, 8

    @inner:
      shl al, 1; jnc @_; mov [edi], Ch
      @_: lea edi, edi+1
      dec cl; jg @inner
    dec ebx; jg @@Loop

    pop eax

  @@End: pop ebx; pop edi; pop esi
  @@Stop:
end;

function getBitsWide(const I: int64; const preferedBitsWide: integer): integer; assembler asm
    mov ecx, preferedBitsWide
    mov eax, 64
    test  I.r64.hi, -1; jnz @@check
    shr eax, 1 // 32
    mov edx, I.r64.lo
    test edx, 0ffff0000h; jnz @@check
    shr eax, 1 // 16
    test edx, 0ff00h; jnz @@check
    shr eax, 1 // 8
    test edx, 0f0h; jnz @@check
    shr eax, 1 // 4
    test edx, 1100b; jnz @@check
    shr eax, 1 // 2
    test edx, 10b; jnz @@check
    shr eax, 1 // 1
  @@check: cmp ecx, eax; jl @@end
  @@userSize: mov eax, ecx
  @@end: //and eax, $ff
end;

function octs(const Buffer: pointer; const BufferLength: integer): string; register overload assembler asm
  @@start:
    or eax, eax; jz @@Stop     // insanity checks
    or edx, edx; jg @@Begin    // can not exceed 2GB
    xor eax, eax; jmp @@Stop
  @@Begin: push esi; push edi; push ebx
    mov esi, Buffer
    mov ebx, BufferLength
    mov eax, Result
    call System.@LStrClr       // cleared first to avoid invalid ptr assignment
    lea edx, [ebx*2 + ebx]
    call System.@LStrSetLength
    mov edi, [eax]; push edi

    mov dx, '00'
    mov cx, $707

  @@Loop: lodsb; mov ah, al
      shr al, 6; or al, dl; stosb
      mov al, ah; shr al, 3
      and ax, cx; or ax, dx; stosw
    dec ebx; jg @@Loop

    pop eax

  @@End: pop ebx; pop edi; pop esi
  @@Stop:
end;

function octs(const I: Int64): string; overload; asm
  mov ecx, I.r64.lo
  mov edx, I.r64.hi
  xchg cl, ch; rol ecx, 16; xchg cl, ch
  xchg dl, dh; rol edx, 16; xchg dl, dh
  push ecx; push edx
  mov ecx, @Result // @Result is currently in eax
  mov edx, 8; mov eax, esp
  call octs
  pop ecx; pop edx
end;

function octs(const I: integer): string; overload; asm
  xchg al, ah; rol eax, 16; xchg al, ah
  push eax
  mov ecx, @Result  // @Result is currently in edx
  mov edx, 4; mov eax, esp
  call octs
  add esp, 4
end;

function bins(const Buffer: pointer; const BufferLength: integer): string; register overload assembler asm
  @@Start:
    or eax, eax; jz @@Stop     // insanity checks
    or edx, edx; jg @@Begin    // can not exceed 2GB
    xor eax, eax; jmp @@Stop
  @@Begin: push esi; push edi; push ebx
    mov esi, Buffer
    // edx (as well as ecx), usually destroyed by Str operation
    mov ebx, edx               // BufLen
    mov eax, Result            // where the result will be stored
    call System.@LStrClr       // cleared first to allocate new string
    //shl edx, 3
    lea edx, [ebx*8]           // BufferLength * 8 bits
    call System.@LStrSetLength // result: new allocation pointer in EAX
    mov edi, [eax]               // put new allocated ptr of @Result to edi

    push edi

    lea ecx, [ebx*2]
    mov eax, '0000'; rep stosd

    mov edi, [esp]
    //mov ecx, ebx               // get bufferlength back to ecx
    mov Ch, '1'                  // buffer fillin value

  @@Loop: lodsb; mov cl, 8

      @inner:
      shl al, 1; jnc @_; mov [edi], Ch
      @_: lea edi, edi+1; dec cl; jg @inner
      dec ebx; jg @@Loop

    pop eax

  @@End: pop ebx; pop edi; pop esi
  @@Stop:
end;

function bins(const I: Int64; const BitsWide: integer = 32): string;
const
  Ch: array[boolean] of Char = ('0', '1');
  //BitsPerByte = 8;
  //ByteSize = Sizeof(byte) * BitsPerByte; // 8
  //WordSize = Sizeof(WORD) * BitsPerByte; // 16
  //LongWordSize = Sizeof(Longword) * BitsPerByte; //32
  //Int64Size = Sizeof(Int64) * BitsPerByte; //64

var
  n, r: integer;
begin
  //if i <= high(byte) then n := ByteSize //Sizeof(byte) * BitsPerByte
  //else if i <= high(WORD) then n := WordSize //Sizeof(WORD) * BitsPerByte
  //else if i <= high(Longword) then n := LongWordSize //Sizeof(Longword) * BitsPerByte
  //else n := Int64Size; //Sizeof(Int64) * BitsPerByte;
  // //if (NumOfBits > n) and (NumOfBits in [Sizeof(byte) * BitsPerByte, Sizeof(WORD) * BitsPerByte,
  // // Sizeof(Longword) * BitsPerByte, Sizeof(Int64) * BitsPerByte]) then n := NumOfBits;
  //if (NumOfBits > n) and ((NumOfBits = ByteSize) or (NumOfBits = WordSize) or
  //  (NumOfBits = LongWordSize) or (NumOfBits = Int64Size)) then n := NumOfBits;
  //Result := '';
  //for n := 0 to n - 1 do
  //  Result := Ch[boolean(i shr n and 1)] + Result;

  n := getBitsWide(I, BitsWide);
  setlength(Result, n);
  fillchar(Result[1], n, '0');
  for r := n - 1 downto 0 do
    if (r < 65) and boolean(i shr r and 1) then
      Result[n - r] := '1';
end;

const
  // we should not use long string here
  // because the long string is really a pointer
  // (storage allocated elsewhere)
  HEX_UPPERCASE: string[32] = '0123456789ABCDEF';
  HEX_LOWERCASE: string[32] = '0123456789abcdef';

// when the caller allocating the pointer @Result,
// do not forget that it might make use of unused register
// (some optimization-rules), in this Hexs-integer function case,
// it would be passed in ecx, believe it?
// this must be taken into account, since it requires a hidden
// yet subtle different handling of String result
//
// as expressed further in the codes below, we dont have to PUSH
// the @Result and manually compute & access it from the stack
// its also not necessary to call UniqueString (the compiler have
// already done it when we set the length by calling LStrSetLength)
//
// the difference comes when the pointer @Result passed by caller
// in the register versus via stack
//
// when the @Result passed in register, then the value returned by
// SetLength (reallocated string pointer) is the (pointer to) string
// itself, we could simply move and process it any further.
//
// when the @Result passed in stack, then the value returned by
// SetLength is the String variable, we have to dereference it
// to be processed later.
// This case occurs in Hexs for int64 and Buffer, since no more
// available register can be used by the compiler, it using stack
// to pass the @Result
//
// other consideration:
// maintaining String reference count is such a daunting task,
// the easier way is simply clearing Result before using it
// (if the Result already empty it does nothing, else it will takes
// care of previous instance of the string, check ref-count, whether
// it was str constant etc., and then freeing memory if necessary)
//
// afterwise then we should call Uniquestring (already done if we are
// explicitly set the length of the string/Result)
//
// btw, don't bother with eax. only the @Result we have to take care about
// it is nice to have eax is also containing @Result, but apparently
// that is not mandatory (at least upon interacting with delphi std lib)
// but beware that in delphi help, eax must contain Result pointer
// see topic: "Assembler procedures and functions"
//

function Hexs(const byte: byte; const uppercase: boolean = YES): string;
const ByteDigits = 2;
asm
  //PUSH Result               //@Result = ECX
  PUSH ebx
  lea ebx, HEX_UPPERCASE +1 // init ebx with default value
  test uppercase, 1         // is uppercase flag = YES?
    jne @skipsetlocase      // nz = YES, then dont bother to kowercase
  lea ebx, HEX_LOWERCASE +1 // uppercase flags = FALSE
  @skipsetlocase:

  mov ah, al; push eax      // save copy first

  mov eax, Result           // where the result will be stored
  call System.@LStrClr      // cleared for ease
  mov edx, ByteDigits       // how much length of str requested
  call System.@LStrSetLength// result: new allocation pointer in EAX
  mov Result, [eax]         // eax contains the new allocated pointer
                            // we got the storage as well at once
  pop eax

  and al, $f; xlat          // strip high nibbles, translate
  mov Result+1, al             // save to allocated string / last position
  mov al, ah
  shr eax, 3 * 4            // shift 3 nibbles away (high nibble of AH)
  and al, $f; xlat          // strip high nibbles, translate
  mov [Result], al             // save to allocated string / first position

  //mov Result, edx
  //call System.@LStrLAsg
  pop ebx
  mov eax, @Result
end;

function Hexs(const word: word; const uppercase: boolean = YES): string; overload;
const WordDigits = 4;
asm
  //PUSH Result               //@Result = ECX
  PUSH ebx
  lea ebx, HEX_UPPERCASE +1 // init ebx with default value
  test uppercase, 1         // is uppercase flag = YES?
    jne @skipsetlocase      // nz = YES, then dont bother to kowercase
  lea ebx, HEX_LOWERCASE +1 // uppercase flags = FALSE
  @skipsetlocase:

  //xchg al, ah; push ax, shl eax, 16; pop ax
  mov dl, ah; mov dh, al;
  mov ax, dx; shl eax, 16; mov ax, dx

  PUSH eax

  mov eax, Result           // where the result will be stored
  call System.@LStrClr      // cleared for ease
  mov edx, WordDigits       // how much length of str requested
  call System.@LStrSetLength// result: new allocation pointer in EAX system
  mov Result, [eax]            // eax contains the new allocated pointer
                            // we got the storage as well at once

  pop eax                   // al

  mov al, ah                // get the high byte
  and al, $f; xlat          // strip high nibbles, translate
  mov Result+3, al          // save to allocated string / last position

  mov al, ah; shr al, 4     // reget the high byte; get high nibble
  and al, $f; xlat          // strip high nibbles, translate
  mov Result+2, al          // save to allocated string / first position

  shr eax, 16               //
  mov ah, al                // we dont again need ah

  and al, $f; xlat
  mov Result+1, al

  mov al, ah
  shr al, 4
  and al, $f; xlat
  mov [Result], al

  pop ebx
  mov eax, @Result
end;

function Hexs(const integer: integer; const uppercase: boolean = YES): string; overload;
const
  IntDigits = 8;
  bswap_eax = $C80F;
begin //@Result = ECX
  asm
    PUSH ebx
    lea ebx, HEX_UPPERCASE+1  // init ebx with default value
    test uppercase, 1;        // is uppercase flag = YES?
      jne @skipsetlocase      // nz = YES, then dont bother to kowercase
    lea ebx, HEX_LOWERCASE+1  // uppercase flags = FALSE
    @skipsetlocase:

    db $f,$c8                 // bswap eax; $c9=ecx $ca=edx $cb=ebx
    mov integer, eax

    mov eax, Result           // where the result will be stored
    call System.@LStrClr      // cleared for ease
    mov edx, IntDigits        // how much length of str requested
    call System.@LStrSetLength// result: new allocation pointer in EAX
    mov edx, [eax]            // eax contains the new allocated pointer
                              // we got the storage as well at once

    mov eax, integer
    mov ecx, IntDigits / 2
    mov ch, al                // initialize

  @Loop:
    shr al, 4
    and al, $f; xlat
    mov [edx], al
    lea edx, edx +1

    mov al, ch
    and al, $f; xlat
    mov [edx], al
    lea edx, edx +1

    shr eax, 8
    mov ch, al
    dec cl
    jnz @Loop

    pop ebx
    mov eax, @Result
  end;
end;

function Hexs(const I: int64; const uppercase: boolean = YES): string; overload;
const I64digits = 16;
begin
  asm //@Result in stack
    PUSH ebx
    lea ebx, HEX_UPPERCASE +1 // init ebx with default value
    test uppercase, 1         // is uppercase flag = YES?
      jne @skipsetlocase      // nz = YES, then dont bother to kowercase
    lea ebx, HEX_LOWERCASE +1 // uppercase flags = FALSE
    @skipsetlocase:

    mov eax, I.r64.lo; db $f,$c8   // bswap eax
    mov I.r64.lo, eax

    mov eax, I.r64.hi; db $f,$c8   // bswap eax
    mov I.r64.hi, eax

    mov eax, Result           // where the result will be stored
    call System.@LStrClr      // cleared for ease
    mov edx, I64Digits        // how much length of str requested
    call System.@LStrSetLength// result: new allocation pointer in EAX
    mov edx, [eax]            // eax contains the new allocated pointer -Differ-

    mov eax, I.r64.hi
    mov ecx, 4; mov ch, al

  @Loop1:
    shr al, 4
    and al, $f; xlat
    mov [edx], al
    lea edx, edx +1

    mov al, ch; //shr al, 4
    and al, $f; xlat
    mov [edx], al
    lea edx, edx +1

    shr eax, 8
    mov ch, al
    dec cl
    jnz @Loop1

    mov eax, I.r64.lo
    mov ecx, 4; mov ch, al

  @Loop2:
    shr al, 4
    and al, $f; xlat
    mov [edx], al
    lea edx, edx +1

    mov al, ch; //shr al, 4
    and al, $f; xlat
    mov [edx], al
    lea edx, edx +1

    shr eax, 8
    mov ch, al
    dec cl
    jnz @Loop2

    pop ebx
    mov eax, @Result
  end;
end;

function Hexs(const Buffer: pointer; const BufferLength: integer;
  const Delimiter: Char = #0; const Uppercase: boolean = YES): string;
asm //@Result in stack
  @@Start:
    or eax, eax; jz @@Stop     // insanity checks
    or edx, edx; jg @@Begin//jle @@Stop
    xor eax, eax
    jmp @@Stop

  @@Begin:
    PUSH esi; PUSH edi
    mov esi, buffer
    mov edi, bufferlength      // save buflength first!

    shl edx, 1                 // edx = Bufferlength * 2
    and ecx, $ff               // note: ecx IS delimiter
    PUSH ecx                   // save it, will be destroyed by LStrSetLength

    cmp ecx, 0; jz @nolimit    // if delim = #0 then skip increase length
    lea edx, edx+edi           // ~> inc(edx, edi)
    dec edx                    // we don't need trailing delimiter

  @nolimit:
    mov eax, Result            // where the result will be stored
    push edx
    call System.@LStrClr       // cleared for ease
    pop edx
    call System.@LStrSetLength // result: new allocation pointer in EAX
    //call System.UniqueString;
                               // ecx, edx destroyed

    mov ecx, edi               // get bufferlength back

    //wrong!mov edi, eax               // WRONG! eax contains the new allocated pointer
                               // we got the storage as well at once
    mov edi, [eax]             // eax contains the new allocated pointer -Differ-

    pop edx                    // get delimiter back

    PUSH ebx;
    lea ebx, HEX_UPPERCASE +1  // get Translation Table
    test Uppercase, 1; jne @skipsetlocase  // is uppercase flag = YES?
    lea ebx, HEX_LOWERCASE +1 // uppercase flags = FALSE

  @skipsetlocase:

    cmp edx, 0
    Jz @@WithoutDelimiter

    dec ecx
    jz @LastByte

    @@WithDelimiter:
      lodsb                    // load byte to AL
      mov ah, al               // copy to AH for second nibble translation

      shr al, 4                // extract high nibble
      xlat; mov [edi], al      // translate and store result
      mov al, ah; and al, $f   // extract low nibble; validate
      xlat; mov edi+1, al      // translate and store result
      mov edi+2, dl            // put delimiter
      lea edi, edi+3           // inc edi by 3
      dec ecx; jg @@WithDelimiter

   @LastByte:
      lodsb                    // load byte to AL
      mov ah, al               // copy to AH for second nibble translation

      shr al, 4                // extract high nibble
      xlat; mov [edi], al      // translate and store result
      mov al, ah; and al, $f   // extract low nibble; validate
      xlat; mov edi+1, al      // translate and store result
      //we donot need these anymore...
      //lea edi, edi+3           // inc edi by 3
      //mov edi+2, dl            // put delimiter
      //dec ecx; jg @@WithDelimiter
      jmp @@Done

    @@WithoutDelimiter:
      lodsb                    // load byte to AL
      mov ah, al               // copy to AH for second nibble translation

      shr al, 4                // extract high nibble
      xlat; mov [edi], al      // translate and store result
      mov al, ah; and al, $f   // extract low nibble; validate
      xlat; mov edi+1, al      // translate and store result
      //mov edi+2, dl            // put delimiter
      //lea edi, edi+3           // inc edi by 3
      lea edi, edi+2           // inc edi only 2 in this block
      dec ecx; jg @@WithoutDelimiter
      jmp @@Done

    @@Done:pop ebx; pop edi; pop esi
    mov eax, @Result
  @@Stop:
end;

function Hexs(const Buffer: pointer; const BufferLength: integer;
  const Uppercase: boolean; const Delimiter: Char = #0): string; overload;
begin
  Result := Hexs(Buffer, BufferLength, Delimiter, Uppercase);
end;

function setBit(const BitNo, I: integer): integer; overload;
begin
  Result := I or (1 shl BitNo);
end;

function isBitSet(const BitNo, I: integer): Boolean; overload;
begin
  Result := (I and (1 shl BitNo)) <> 0;
end;

function ResetBit(const BitNo, I: integer): integer; overload;
begin
  Result := I and ((1 shl BitNo) xor -1);
end;

function ToggleBit(const BitNo, I: integer): integer; overload;
begin
  Result := I xor (1 shl BitNo);
end;

function setBit(const BitNo: integer; const I: int64): int64; overload;
begin
  Result := I or (int64(1) shl BitNo);
end;

function isBitSet(const BitNo: integer; const I: int64): boolean; overload;
begin
  Result := (I and (int64(1) shl BitNo)) <> 0;
end;

function ResetBit(const BitNo: integer; const I: int64): int64; overload;
begin
  Result := I and ((int64(1) shl BitNo) xor int64(-1));
end;

function ToggleBit(const BitNo: integer; const I: int64): int64; overload;
begin
  Result := I xor (int64(1) shl BitNo);
end;

// ~~~~~~~~~~~~~~~~~~~~~~~
// Pseudo-random generator
// ~~~~~~~~~~~~~~~~~~~~~~~
var
  factors: array[boolean] of int64 = ($13FB7DD4FFC7, 7627861919189 - 1);
  X: integer absolute RandSeedEx;
  f: integer absolute factors;

function RandShuffle: int64; register asm
  push edi

  mov eax, f.0; mul X.16              // x4
  mov ecx, eax; mov eax, X.12         // x3
    mov edi, edx
  mov X.16, eax; mul f.8
  add ecx,eax; mov eax, X.8           // x2
    adc edi, edx
  mov X.12, eax; mul f.12
  add ecx, eax; mov eax, X.0          // x0
    adc edi, edx
  mov X.8, eax; mul f.4
    add eax, ecx; adc edx, edi
    add eax, X.4; adc edx, 0
  mov X.0, eax; mov X.4, edx          // x1

  pop edi
end;

function RandomInt64: int64; register asm call RandShuffle end;
function RandomInt(const Max: cardinal): cardinal; register asm
  test Max, -1; jnz @begin; ret
@begin: push Max
  call RandShuffle; pop edx
  mul edx; mov eax, edx
end;

function RandomInt(const Min, Max: integer): integer; register asm
// Min-Max range should not exceed cardinal boundary minus 1
// (max difference = 4294967295)
  sub max, min; jns @_
    xor eax, eax; ret  // zeronize
  @_: inc max          // difference = (max - min) +1
  push min; push max   // ...save
  call RandShuffle     // get R
  pop edx; pop ecx     // ..restore
  mul edx              // multiply R by difference (truncated)
  lea eax, edx + ecx
end;

type
  f80 = packed record // 80 bits extended floating point
    lo, hi: integer;
    exp: word;
  end;

function RandomExt: Extended; register asm
  call RandShuffle     //  edx:eax
  or edx, 1 shl 31     //  normalized bit-63 in edx
  mov Result.f80.lo, eax
  mov Result.f80.hi, edx
  mov Result.f80.exp, 3fffh
  fLd1                 //  load 1.0, since...
  fLd Result           //  1 < Result < 2
  fSubRP               //  after sub, now 0 < Result < 1, pop the stack
  fStP Result          //  store back with pop
  wait
end;

{ Original version
  RandomDbl PROC NEAR
  public RandomDbl
    CALL    RandomBit            ; random bits
    MOV     EDX, EAX             ; fast conversion to float
    SHR     EAX, 12
    OR      EAX, 3FF00000H
    SHL     EDX, 20
    MOV     DWORD PTR [TEMP+4], EAX
    MOV     DWORD PTR [TEMP], EDX
    FLD1
    FLD     QWORD PTR [TEMP]     ; partial memory stall here
    FSUBR
    RET
  RandomDbl ENDP
}

procedure RandomInit(const I: integer = __AAMAGIC0__); register
const
  PRIME0 = 7;
  PRIME1 = $01C8E80D; // 29943821 : 1 1100 1000 1110 1000 0000 1101 ~ 1 11001000 11101000 00010101
  e: extended = 0; // use extended to allow broader range of generated number
asm
  push edi; mov edi, 4
  @LFill:
    imul eax, PRIME1
    dec eax; mov X[edi*4], eax
    dec edi; jge @LFill

  mov edi, PRIME0
  @@LRand: call RandomExt
    fstp e // note: for consistency, never directly alter X in RandomExt
    mov eax, e.f80.lo; xor X, eax
    //warn: we should not use edx, since it has been convoluted.
    //(the highest bit is always 1)
    //mov edx, e.ext.hi; mov X.4, edx
    dec edi; jnz @@LRand
  pop edi
end;

//const
//  BitsperByte = 8;
//  BitsPerInt = SizeOf(integer) * BitsPerByte;

type
  TBitEnum = 0..BitsPerInt - 1;
  TBitSet = set of TBitEnum;
  PBitArray = ^TBitArray;
  TBitArray = array[0..4096] of TBitSet;

destructor TBits.Destroy;
begin
  setSize(0);
  inherited Destroy;
end;

procedure TBits.Error;
begin
  raise Self.Create;
end;

procedure TBits.setSize(Value: integer);
var
  NewMem: Pointer;
  NewMemSize: integer;
  OldMemSize: integer;

  function Min(X, Y: integer): integer;
  begin
    Result := X;
    if X > Y then
      Result := Y;
  end;

begin
  if Value <> Size then begin
    if Value < 0 then
      Error;
    NewMemSize := ((Value + BitsPerInt - 1) div BitsPerInt) * SizeOf(integer);
    OldMemSize := ((Size + BitsPerInt - 1) div BitsPerInt) * SizeOf(integer);
    if NewMemSize <> OldMemSize then begin
      NewMem := nil;
      if NewMemSize <> 0 then begin
        GetMem(NewMem, NewMemSize);
        FillChar(NewMem^, NewMemSize, 0);
      end;
      if OldMemSize <> 0 then begin
        if NewMem <> nil then
          Move(fBits^, NewMem^, Min(OldMemSize, NewMemSize));
        FreeMem(fBits, OldMemSize);
      end;
      fBits := NewMem;
    end;
    fSize := Value;
  end;
end;

procedure TBits.setBit(Index: integer; Value: Boolean); assembler;
asm
  cmp Index, [eax].fSize
  jae @@exSize
@@proc: mov eax, [eax].fBits
  or Value, Value; jz @@test&Reset
  bts [eax], Index; RET
@@test&Reset: btr [eax], Index; RET
@@exSize: cmp Index, 0; jl TBits.Error
  push eax {Self}; push Index; push ecx {Value}
  inc Index; CALL TBits.setSize
  pop ecx {Value}; pop Index; pop Self; jmp @@proc
end;

function TBits.getBit(Index: integer): Boolean; assembler;
asm
  cmp Index, [eax].fSize; jae TBits.Error
  mov eax, [eax].fBits
  bt [eax], Index; sbb eax, eax
  and eax, 1
end;

function TBits.OpenBit: integer;
var
  i: integer;
  b: TBitSet;
  e: TBitEnum;
  n: integer;
begin
  n := (Size + BitsPerInt - 1) div BitsPerInt - 1;
  for i := 0 to n do
    if PBitArray(fBits)^[i] <> [0..BitsPerInt - 1] then begin
      b := PBitArray(fBits)^[i];
      for e := Low(e) to High(e) do begin
        if not (e in b) then begin
          Result := i * BitsPerInt + e;
          if Result >= Size then
            Result := Size;
          Exit;
        end;
      end;
    end;
  Result := Size;
end;

destructor TIDStr.Destroy;
begin
  Count := 0;
end;

constructor TIDStr.Create;
begin
  Create(0);
end;

constructor TIDStr.Create(const Size: integer);
begin
  inherited Create;
  Count := Size;
end;

function TIDStr.delete(const ID: integer): boolean;
begin
  Result := FALSE;
end;

procedure TIDStr.join(const IntStr: TIDStr);
begin
end;

procedure TIDStr.Clear;
begin
  Count := 0;
end;

function TIDStr.getValue(const ID: integer): string;
var
  i: integer;
begin
  if (fCount < 1) then
    Result := ''
  else if ID = fMaxID then
    Result := fStrs[fCount - 1]
  else begin
    i := fInts[0] - 1;
    repeat inc(i); until (i >= fCount) or (fInts[i] = ID);
    if i >= fCount then
      Result := ''
    else
      Result := fStrs[i]
  end;
end;

procedure TIDStr.setValue(const ID: integer; const Value: string);
var
  i: integer;
begin
  if ID < 0 then
    raise TIDStr.Create at@ID;
  if fMaxID < ID then begin
    fMaxID := ID;
    i := fCount;
    inc(fCount);
    SetLength(fInts, fCount); SetLength(fStrs, fCount);
  end
  else begin
    i := -1;
    repeat inc(i); until (fInts[i] = ID);
  end;
  fInts[i] := ID;
  fStrs[i] := Value;
end;

function TIDStr.IDOf(const Value: string): integer;
var
  i: integer;
begin
  Result := -1;
  for i := 0 to fCount - 1 do
    if fStrs[i] = Value then begin
      Result := fInts[i];
      break;
    end;
end;

function TIDStr.IDAtPos(const Pos: integer): integer;
begin
  if Pos >= fCount then
    Result := -1
  else
    Result := fInts[Pos];
end;

function TIDStr.ValueAtPos(const Pos: integer): string;
begin
  if Pos >= fCount then
    Result := ''
  else
    Result := fStrs[Pos];
end;

procedure TIDStr.setCount(const Value: integer);
begin
  if fCount < Value then begin
    fCount := Value;
    SetLength(fInts, Value);
    SetLength(fStrs, Value);
  end;
  if fCount = 0 then
    fMaxID := -1
  else
    fMaxID := fInts[fCount - 1];
end;

end.

